//#define ATMEGA168_PLATTFORM
//#define ARDUINO_PLATFORM												//bootloader development on arduino plattform
#define PINGPONG_PLATTFORM

#ifdef PINGPONG_PLATTFORM
#define ATMEGA8_PLATTFORM
#endif


#ifdef ARDUINO_PLATFORM
#define ARDUINO_DEBUG
#endif

/*
	AudioBoot - flashing a microcontroller by PC line out 
				this verision is for ping pong game

	Hardware: 	Atmega168 
				Atmega8

	input pin: 	should be connected to a symmetric voltage divider with two 10K resistors,
				the input pin is also connected by a 100nF capacitor to the PC line out

	outuput: 	status led connected to a pin with a 470 Ohm resistor


	As developement platform an Arduino Diecimilla was used. Therefore you
	will find many #ifdefs for the Arduino in this code.
	If you want to optimize the bootloader further you may use an Arduino 
	as developement platform.


	necessary setup

	1. Project->ConfigurationOptions->Processortype
	2. Project->ConfigurationOptions->Programming Modell 'Os'
	3. Project->ConfigurationOptions->CustomOptions->LinkerOptions->see further down
	5. BootresetvectorFuseBit enable Bootloader
	6. compile and flash Bootloader




	ATMEGA168	(1Kbyte bootloader) Fuses: BOOTSz1=0(x);BOOTSz0=1(o); Linker: -Wl,--section-start=.text=0x3c00

	Fuses Atmega168: 
	Extended: 	0xFA
	HIGH: 		0xDF
	LOW: 		0xFF

	ATMEGA8		(1Kbyte bootloader) Fuses: BOOTSz1=0(x);BOOTSz0=1(o); Linker: -Wl,--section-start=.text=0x1c00

	Fuses for PingPong, Atmega8: 
	HIGH:		0xDA
	LOW:		0xE4
	LOCK: 		0xCF : LPM and SPM prohibited in bootloader section




	(c) -C-H-R-I-S-T-O-P-H-   -H-A-B-E-R-E-R- 2011

	v0.1	19.6.2008	C. -H-A-B-E-R-E-R- 	Bootloader for IR-Interface
	v1.0	03.9.2011	C. -H-A-B-E-R-E-R-	Bootloader for audio signal
	v1.1	05.9.2011	C. -H-A-B-E-R-E-R-	changing pin setup, comments, and exitcounter=3 
	v1.2	12.5.2012	C. -H-A-B-E-R-E-R-	Atmega8 Support added, java programm has to be addeptet too 
	v1.3	20.5.2012	C. -H-A-B-E-R-E-R-	now interrupts of user programm are working 
	v1.31	27.5.2012	25mmHg				2LEDs (PB0,PB1) on PingPongBoard are now working for AudioLevelCheck on PD0 (is AudioInput)

	This program is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
 	the Free Software Foundation; either version 2 of the License, or
 	(at your option) any later version.
*/
//***************************************************************************************
//** the data transmition was developed on the arduino plattform
//***************************************************************************************
#ifdef ARDUINO_PLATFORM

	const int ledPin =  13;												//LED connected to digital pin 13

	void setup()
	{                
	  pinMode(ledPin, OUTPUT);											//initialize the digital pin as an output   
	  Serial.begin(9600); 
      Serial.println("AudioBoot starting");
	}

	void loop()
	{
	  a_main();
	  while(1); // stop
	}
	
	#define LEDON			{digitalWrite(ledPin, HIGH);}
	#define LEDOFF			{digitalWrite(ledPin, LOW );}
    #define TOGGLELED		{digitalWrite(ledPin, !digitalRead(ledPin));} 

	#define INPUTAUDIOPIN	(1<<PB4)
	#define PINVALUE		(PINB&INPUTAUDIOPIN)

	#define PINLOW			(PINVALUE==0)
	#define PINHIGH			(!PINLOW)

#endif
//***************************************************************************************
// end of arduino specific developement plattform code
//***************************************************************************************


//***************************************************************************************
// here starts the code for the stand alone microcontroller
//***************************************************************************************
#ifdef ATMEGA168_PLATTFORM

	#include <avr/io.h>
	#include <avr/interrupt.h>
	#include <avr/signal.h>
	#include <stdlib.h>
	#include <avr/boot.h>

	#define LEDPORT 		(1<<PB5);
	#define INITLED			{DDRB|=LEDPORT;}
	#define LEDON			{PORTB|=LEDPORT;}
	#define LEDOFF			{PORTB&=~LEDPORT;}
	#define TOGGLELED		{PORTB^=LEDPORT;}

	// use TxD ( PD1 ) as audio input	
	#define INPUTAUDIOPIN	(1<<PD1)
	#define PINVALUE		(PIND&INPUTAUDIOPIN)

	// turn on pull up 
	#define INITPORT		{PORTD|=INPUTAUDIOPIN;} 

	#define PINLOW			(PINVALUE==0)
	#define PINHIGH			(!PINLOW)

	#define true			(1==1)
	#define false			!true

	int main()
	{
		INITLED;
		INITPORT; 
		a_main(); // start the main function
	}

#endif

#ifdef PINGPONG_PLATTFORM

	#include <avr/io.h>
	#include <avr/interrupt.h>
	#include <avr/signal.h>
	#include <stdlib.h>
	#include <stdbool.h>
	#include <avr/boot.h>

	// use RxD ( PD0 ) as audio input and PD2 as PWRGND for Poti
	#define INPUTAUDIOPIN	(1<<PD0)
	//#define PINVALUE (PIND&INPUTAUDIOPIN)								//is now pinRead() look at line 195
	#define PINVALUE		pinRead()
	#define POTIGNDPIN		(1<<PD2)
	
	// turn on pull up and power on Poti
	#define INITPORT 		{PORTD|=INPUTAUDIOPIN; DDRD |=POTIGNDPIN;}	//set Poti GND and set PullUp


	#define PINLOW			(PINVALUE==0)
	#define PINHIGH			(!PINLOW)

	// ping pong led definitions
	#define LEDPORT 		((1<<PB0)|(1<<PB1))							//now 2LEDs 
	#define INITLED 		{DDRB|=0x3; DDRC|=0x0F; DDRD|=0xF0;}

	#define LEDHIGH			{PORTB|=(1<<PB0); PORTB&=~(1<<PB1);}			//now HIGH-Level on INPUTAUDIOPIN
	#define LEDLOW			{PORTB|=(1<<PB1); PORTB&=~(1<<PB0);}			//now LOW-Level on INPUTAUDIOPIN
	#define LEDON			{PORTB|= LEDPORT;}
	#define LEDOFF			{PORTB&=~LEDPORT;}
	//#define TOGGLELED		{PORTB^=LEDPORT;}							//old 
	#define TOGGLELED		{if(PORTB&LEDPORT)LEDOFF else LEDON}		//now 2 LEDs ready for Toggeling
	
	#define LATCHCLKPIN		(1<<PB2)
	#define SDATAPIN		(1<<PB4)
	#define RCLKPIN			(1<<PB3)

	#define LATCHCLK		{PORTB|=LATCHCLKPIN;}
	#define SDATA			{PORTB|=SDATAPIN;}
	#define RCLK 			{PORTB|=RCLKPIN;}
	#define SETDATA			(PORTB|SDATA)

	#define RESETALL		{PORTB&=~(SDATAPIN|RCLKPIN|LATCHCLKPIN);}
	#define INITLATCH		{DDRB|=(SDATAPIN|RCLKPIN|LATCHCLKPIN);}

uint8_t pinRead()
{
if (PIND&INPUTAUDIOPIN) LEDHIGH 
					else LEDLOW 
return (PIND&INPUTAUDIOPIN);											//now returns the same as the code on line 161
}

void setSpalte(uint8_t k)
{
	uint8_t value;
	value=PORTC;
	value&=~0x0F;
	value|=k&0x0F;
	PORTC=value;

	value=PORTD;
	value&=~0xF0;
	value|=k&0xF0;
	PORTD=value;
}

/****************************************************************************************

	void SetExternalPortByte(uint8_t port,uint8_t data)

	Move data to external shift register.
	
	common pins: data, latchout
	single pin: shift register clock

	22.12.2007 ch

****************************************************************************************/
void setExternalPort(uint16_t data)
{
	uint8_t i;

	RESETALL;

	data=~data;

	for (i=0; i<12; i++)
	{
  		if(data & (1<<11)) SDATA;

  		RCLK;
		data= data << 1;

		RESETALL;
	}
	LATCHCLK;															// clock all latches data to output
	RESETALL;
}

	int main()
	{		
		INITLED;
		INITLATCH;
		INITPORT; 
		setExternalPort(1);
		a_main();														// start the main function
	}

#endif

//***************************************************************************************
// main loop
//***************************************************************************************
#define TIMER TCNT2														// we use timer2 for measuring time

// frame format definition
#define COMMAND				0
#define PAGEINDEXLOW 		1  // page address lower part
#define PAGEINDEXHIGH 		2  // page address higher part
#define CRCLOW				3  // checksum lower part 
#define CRCHIGH 			4  // checksum higher part 
#define DATAPAGESTART		5  // start of data

#ifdef 	ATMEGA168_PLATTFORM
#define PAGESIZE 128
#define FRAMESIZE			(PAGESIZE+DATAPAGESTART)					// size of the data block to be received = 128+5 = 133
#endif

#ifdef 	PINGPONG_PLATTFORM
#define PAGESIZE 64
#define FRAMESIZE			(PAGESIZE*2+DATAPAGESTART)					// size of the data block to be received = 2x(64)+5 = 133
#endif

// bootloader commands
#define NOCOMMAND			0
#define TESTCOMMAND			1
#define PROGCOMMAND			2
#define RUNCOMMAND			3

// led
#define WAITBLINKTIME		10000

uint8_t FrameData[FRAMESIZE];



//***************************************************************************************
// receiveFrame()
//
// This routine receives a manchester coded signal at the input pin.
// The routine waits for a toggling voltage level. 
// It automatically detects the transmission speed.
//
// output: 		uint8_t flag: true: checksum ok
//				Data // global variable
// 
//***************************************************************************************
uint8_t receiveFrame()
{
  uint16_t store[16];

  uint16_t counter=0;
  uint16_t time=0;
  uint16_t delayTime;
  uint8_t p;
  uint8_t k=8;
  uint8_t dataPointer=0;
  uint16_t n;

  //******************* synchronisation and bit rate estimation *************************
  time=0;

  while(PINLOW);

  TIMER=0; 				// reset timer
  for(n=0;n<16;n++)
  {
    // wait for falling edge
	while(PINHIGH);	
    while(PINLOW);
    p=TIMER;
    store[counter++]=p;
    TIMER=0; 			// reset timer
    if(n>=8)time+=p; 	// time accumulator for mean period calculation only the last 8 times are used 
  }
  
  delayTime=time*3/4/8;

  // wait for rising edge
  while(PINLOW);
  while(PINHIGH);
  TIMER=0;
  p=1;

  //************************************* wait for start bit ****************************
  while(p!=0) 
  {
    // delay 3/4 bit
    while(TIMER<delayTime);
    // wait for edge
    p=PINVALUE;
    while(p==PINVALUE);
    TIMER=0;
    counter++;
  }

  //************************************** receive data bits ****************************
  k=8;
  for(n=0;n<(FRAMESIZE*8);n++)
  {
    // delay 3/4 bit
    while(TIMER<delayTime);
    // wait for edge ( manchester code )
    p=PINVALUE;
    while(p==PINVALUE);
    TIMER=0;
    counter++;
    
    FrameData[dataPointer]=FrameData[dataPointer]<<1;
    if(p==0) FrameData[dataPointer]|=1;
    k--;
    if(k==0){dataPointer++;k=8;};
  }
    uint16_t crc=(uint16_t)FrameData[CRCLOW]+FrameData[CRCHIGH]*256;

  //***************************************** debug *************************************
  #ifdef ARDUINO_DEBUG
  for(n=0;n<10;n++)
  {
    Serial.println(store[n]);
  }
  Serial.print("mean time:");
  Serial.println(time/8);
  Serial.print("3/4 time:");
  Serial.println(delayTime);
  Serial.print("counter:");
  Serial.println(counter);
  Serial.println("*********************");
  for(n=0;n<FRAMESIZE;n++)
  {
    Serial.print(FrameData[n],HEX);
    Serial.print(" ");

    Serial.println((int)FrameData[n]);
  } 
  Serial.println("*********************");

  Serial.print("COMMAND:");
  Serial.println((int)FrameData[COMMAND]);
  Serial.print("PAGEINDEX:");
  Serial.println((int) FrameData[PAGEINDEXLOW]+256*FrameData[PAGEINDEXHIGH]);
  Serial.print("CRC:");

  Serial.println(crc,HEX);
  Serial.println("*********************");
#endif


  if(crc==0x55AA) return true;
  else return false;
}


//***************************************************************************************
//	void boot_program_page (uint32_t page, uint8_t *buf)
//
//  Erase and flash one page.
//
//  input: 		page address and data to be programmed
// 
//***************************************************************************************
void boot_program_page (uint32_t page, uint8_t *buf)
{
	uint16_t i;
  
    cli();						// disable interrupts

    boot_page_erase (page);
    boot_spm_busy_wait ();		// Wait until the memory is erased.

    for (i=0; i<SPM_PAGESIZE; i+=2)
    {
        uint16_t w = *buf++;
        w += (*buf++) << 8;
        
        boot_page_fill (page + i, w);
    	boot_spm_busy_wait();	// Wait until the memory is written.
	}

    boot_page_write (page);		// Store buffer in flash page.
    boot_spm_busy_wait();		// Wait until the memory is written.
	
    boot_rww_enable ();
}


void initstart()
{
   // Timer 2 normal mode, clk/8, count up from 0 to 255  
   // ==> frequency @16MHz= 16MHz/8/256=7812.5Hz

#ifdef ATMEGA168_PLATTFORM
   TCCR2B= _BV(CS21);
#endif

#ifdef ATMEGA8_PLATTFORM
	TCCR2= _BV(CS21);
#endif

}
void runProgramm(void)
{
	// reintialize registers to default
	DDRB=0;
	DDRC=0;
	DDRD=0;
	TCCR2=0;			// turn off timer2
	cli();
	// make sure that interrupt vectors point to user space
	GICR=(1<<IVSEL);
	GICR=0;
	// start user programm
	asm volatile(		
	"clr r30	\n\t"	//load z register with address
	"clr r31	\n\t"	//load z register with address
	"ijmp		\n\t"
	);
}

//***************************************************************************************
// main loop
//***************************************************************************************
void a_main()
{
  initstart();

 #ifdef ARDUINO_DEBUG  
  Serial.println("waiting for signal");
 #endif

  uint8_t p=0;
  uint16_t time=WAITBLINKTIME;
  uint8_t timeout=6;													//timeout for bootloader is 6seconds
 
  //*************** wait for toggling input pin or timeout ******************************
  uint8_t exitcounter=3;
  while(1)
  {  

    if(TIMER>100)  // timedelay ==> frequency @16MHz= 16MHz/8/100=20kHz
    {
       TIMER=0;
       time--;
       if(time==0)
       {
         TOGGLELED;
         setSpalte(1);
		 time=WAITBLINKTIME;
         timeout--;
         if(timeout==0)
         {
         	setSpalte(0);												//now 0 old 2

           LEDOFF; // timeout,
           // leave bootloader and run program
		   runProgramm();
         }
       }
    }
    if(p!= PINVALUE)
    {
	//setSpalte(3);
      p= PINVALUE;
      exitcounter--;
    }
    if(exitcounter==0) break; // signal received, leave this loop and go on
  }
  //*************** start command interpreter *******************************************
  LEDON;
  while(1)
  {
    if(!receiveFrame())
    {
      //*****  error: blink fast, press reset to restart ********************************
      while(1)
      {         
        if(TIMER>100)  // timerstop ==> frequency @16MHz= 16MHz/8/100=20kHz
        {
           TIMER=0;
           time--;
           if(time==0)
           {
             TOGGLELED;
             time=1000;
           }
        }
      }
    }
    else // succeed
    {
      TOGGLELED;

      switch(FrameData[COMMAND])
      {
        case TESTCOMMAND: // not used yet
        {

  #ifdef ARDUINO_DEBUG        
         Serial.println("testcommand");
  #endif

        }
        break;
        case RUNCOMMAND:				//setExternalPort(2);
        {
		
  #ifndef ARDUINO_DEBUG          
		runProgramm();					//leave bootloader and run program
  #endif

  #ifdef ARDUINO_DEBUG        
         Serial.println("runcommand");	//virtual run
  #endif

        }
        break;
        case PROGCOMMAND:
        { 

  #ifdef ATMEGA168_PLATTFORM
  			// Atmega168	Pagesize =64Worte =128Byte
			uint16_t k;
			k=(((uint16_t)FrameData[PAGEINDEXHIGH])<<8)+FrameData[PAGEINDEXLOW];
			boot_program_page (SPM_PAGESIZE*k, FrameData+DATAPAGESTART);					// erase and programm page
  #endif

  #ifdef ATMEGA8_PLATTFORM
  			// Atmega8		Pagesize =32Worte =64Byte
			uint16_t k;
			k=(((uint16_t)FrameData[PAGEINDEXHIGH])<<8)+FrameData[PAGEINDEXLOW];
			boot_program_page (SPM_PAGESIZE*k*2, FrameData+DATAPAGESTART);					// erase and programm page
			boot_program_page (SPM_PAGESIZE*(k*2+1), FrameData+SPM_PAGESIZE+DATAPAGESTART);	// erase and programm page
	
	#ifdef	PINGPONG_PLATTFORM
			setSpalte(k);												//show binary frame number on ping pong game
	#endif

  #endif

  #ifdef ARDUINO_DEBUG 
          Serial.println("programming");
  #endif

        }
        break;
      }
      FrameData[COMMAND]=NOCOMMAND;										//delete command
    }
  }
}
